#include <bsp.h>
#include <bsp/irq.h>
#include <bsp/openpic.h>
#include <libcpu/io.h>
#include <rtems/error.h>
#include <rtems.h>

#include <assert.h>
#include <errno.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <time.h>

#include "semcount.h"

#include "mvme3100_timer.h"


#ifndef BSP_MISC_IRQ_LOWEST_OFFSET
#define BSP_MISC_IRQ_LOWEST_OFFSET BSP_PCI_IRQ_LOWEST_OFFSET
#endif

#ifndef OPENPIC_COUNT_MAX
#define OPENPIC_COUNT_MAX OPENPIC_COUNT_MASK
#endif


static bool mvme3100_timer_allocated[OPENPIC_NUM_TIMERS] = { false };

static uint32_t openpic_timer_Hz = 0;


static int mvme3100_timer_disable(const struct mvme3100_timer *timer);

static void mvme3100_timer_interrupt(void *arg);

static int mvme3100_timer_set(struct mvme3100_timer *timer,
                              uint32_t              openpic_timer_cycles);


static int mvme3100_timer_disable(const struct mvme3100_timer *timer)
{
	assert(timer);
	assert(timer->index < (sizeof (mvme3100_timer_allocated) / sizeof (*mvme3100_timer_allocated)));


	openpic_settimer(timer->index, OPENPIC_TIMER_COUNT_INHIBIT, 0); // This is important: We must set the "Count inhibit" bit for the next programmed timer value (without the bit set) to be latched


	return (0);
}


int mvme3100_timer_init(struct mvme3100_timer *timer)
{
	assert(timer);


	if (!OpenPIC)
	{
		errno = ENXIO;

		return (-1);
	}


	if (!openpic_timer_Hz)
	{
		if (!(openpic_timer_Hz =
#ifdef BSP_OPEN_PIC_BIG_ENDIAN
		                         in_be32((uint32_t *) &OpenPIC->Global.Timer_Frequency)
#else
		                         in_le32((uint32_t *) &OpenPIC->Global.Timer_Frequency)
#endif
		     ))
		{
			errno = ENXIO;

			return (-1);
		}
	}


	for (size_t i = 0; i < sizeof (mvme3100_timer_allocated) / sizeof (*mvme3100_timer_allocated); i++)
	{
		if (!*(mvme3100_timer_allocated + i))
		{
			int                    error;
			rtems_irq_connect_data timer_irq;


			if ((error = rtems_semaphore_create(42, 0, 0, 1, &timer->semaphore)) != RTEMS_SUCCESSFUL)
				rtems_fatal_error_occurred(1);


			*(mvme3100_timer_allocated + i) = true;

			timer->index = i;


			timer_irq.name   = BSP_MISC_IRQ_LOWEST_OFFSET + timer->index;
			timer_irq.hdl    = mvme3100_timer_interrupt;
			timer_irq.handle = timer;
			timer_irq.on     = NULL;
			timer_irq.off    = NULL;
			timer_irq.isOn   = NULL;


			if (BSP_install_rtems_irq_handler(&timer_irq) != 1)
				rtems_fatal_error_occurred(2);


			if (mvme3100_timer_disable(timer) != 0)
				rtems_fatal_error_occurred(3);

			//! \todo De-magic-number
			openpic_maptimer(timer->index, 1 << 0);
			openpic_inittimer(timer->index, 8, OPENPIC_VEC_SOURCE - BSP_PCI_IRQ_LOWEST_OFFSET + timer_irq.name);


			return (0);
		}
	}


	errno = EAGAIN; // "Resource unavailable, try again"

	return (-1);
}


static void mvme3100_timer_interrupt(void *arg)
{
	assert(arg);


	struct mvme3100_timer *timer = (struct mvme3100_timer *) arg;


	(void) mvme3100_timer_disable(timer);


	printk("p");


	if (rtems_semaphore_release(timer->semaphore) != RTEMS_SUCCESSFUL)
		rtems_fatal_error_occurred(4);


	printk("P");
}


int mvme3100_timer_wait(struct mvme3100_timer *timer,
                        clockid_t             clock_id,
                        uint32_t              openpic_timer_cycles)
{
	assert(timer);


	if (!(timer->index < (sizeof (mvme3100_timer_allocated) / sizeof (*mvme3100_timer_allocated))))
	{
		errno = EINVAL;

		return (-1);
	}


	assert(rtems_semaphore_count(timer->semaphore) == 0);


	printk("s");


	// Set and start the timer
	if (mvme3100_timer_set(timer, openpic_timer_cycles) != 0)
		return (-1);


	printk("S");


	if (rtems_semaphore_obtain(timer->semaphore, 0, 0) != RTEMS_SUCCESSFUL)
		rtems_fatal_error_occurred(5);


	printk("_"); // Done


	assert(rtems_semaphore_count(timer->semaphore) == 0);


	return (0);
}


static int mvme3100_timer_set(struct mvme3100_timer *timer,
                              uint32_t              openpic_timer_cycles)
{
	assert(timer);
	assert(timer->index < (sizeof (mvme3100_timer_allocated) / sizeof (*mvme3100_timer_allocated)));


	assert(rtems_semaphore_count(timer->semaphore) == 0);

#ifdef BSP_OPEN_PIC_BIG_ENDIAN
	assert(in_be32((uint32_t *) &OpenPIC->Global.Timer[timer->index]._Vector_Priority.Reg) & OPENPIC_MASK);
#else
	assert(in_le32((uint32_t *) &OpenPIC->Global.Timer[timer->index]._Vector_Priority.Reg) & OPENPIC_MASK);
#endif


	if (openpic_timer_cycles & ~OPENPIC_COUNT_MASK)
	{
		errno = EINVAL;

		return (-1);
	}


	openpic_settimer(timer->index, openpic_timer_cycles, 1);


	return (0);
}

